/**********************************************************************

 Copyright (c) 2010 Jim Greubel, Chaneru inc. All rights reserved.

 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>

#define DEBUG if ( debug ) printf

typedef struct {
   void * next;
   char * name;
   int    trkcnt;
   int  * tracks;
} LIST;

typedef struct {
   void * next;
   char * name;
   char * path;
} ART;

typedef struct {
   char * name;
   char * value;
} TAG;

int debug=1;

FILE * xml, * out;

int  artist_cnt, album_cnt, playlist_cnt, song_cnt, podcast_cnt, track_cnt, link_cnt, art_cnt;
LIST *artists, *albums, *playlists, *songs, *podcasts, *linksx;
int lines, end, pos=0;
ART *album_art, *artist_art, *music_art;
int keys;
TAG key[1000];

char * flds[4];
char dummy[4]="";

LIST * acq_item(char *,int);
int close_lists();
char * decode_url(char*);
int decode_txt(char*);
int done();
int getflds(char*);
int Hex2Bin(char);
int insert_play();
int insert_trk();
int insert_list(LIST*,char*,int,int);
int Scan(char*,char*);

int main(int argc, char *argv[]) {
   int i, f;
   int tracks=0, playlists=0;
   time_t start, stop;
   char * b, c, buf[5000];
   
   time( &start );
   
   getcwd( buf, 500 );
   printf( "Current directory - %s\n", buf );
   for ( i=0; i<strlen(buf); i++ ) if ( buf[i] == '/' || buf[i] == '\\' ) c = buf[i];
   
   i = strlen( buf );
   if ( strncmp( "Roku", &buf[i-4], 4 ) ) {
      printf( "This MUST be executed in the Roku folder !\n" );
      done();
   }	

   sprintf( buf, "iTunes%ciTunes Music Library.xml", c );
   xml = fopen( buf, "r" );
   if ( !xml ) {
      perror( "* Failed to open iTunes Music Library.xml !" );
      done();
   }

   if ( chdir( ".." ) ) {
      perror( "* Failed to change to Roku directory !" );
      done();
   }

	i = mkdir( "iTunes Tables", 0777 );   // MAC
//	i = mkdir( "iTunes Tables" );         // WIN
   if ( i ) {
      perror( "* Failed to create iTunes Tables folder !" );
   }

   if ( chdir( "iTunes Tables" ) ) {
      perror( "* Failed to change to iTunes Tables folder !" );
      done();
   }

   while ( 1 ) {
      b = fgets( buf, 5000, xml );
      if ( !b ) break;
      lines++;
      
      if ( keys && b[0] != '\t' ) continue;

      f = getflds( b );
      switch ( f ) {
         case 0 :
                  if ( tracks && keys ) {
                     insert_trk();
                     keys = 0;
                  }
                  if ( playlists & !strcmp( flds[0], "/dict" )) {
                     keys = 0;
                  }
                  break;
         case 1 :
                  if ( !strcmp( flds[1], "Tracks" ))    { tracks = 1; break; }
                  if ( !strcmp( flds[1], "Playlists" )) { tracks = 0; playlists = 1; break; }
                  if ( tracks && keys ) {
                     i = keys++;
                     key[i].name  = (char*)malloc( strlen(flds[1])+8 );
                     strcpy( key[i].name,  flds[1] );
                     key[i].value = 0;
                  }
                  if ( playlists && keys && !strcmp( flds[1], "Playlist Items" )) {
                     insert_play();
                     keys = 0;
                  }
                  break;
         case 2 :
                  if ( tracks || playlists ) {
                     i = keys++;
                     key[i].name  = (char*)malloc( strlen(flds[1])+8 );
                     key[i].value = (char*)malloc( strlen(flds[2])+20 );
                     strcpy( key[i].name,  flds[1] );
                     strcpy( key[i].value, flds[2] );
                     if ( !strcmp( key[i].name, "Location" )) {
                        key[i].value = decode_url( key[i].value );
                     } else decode_txt( key[i].value );
                  }
                  break;
      }
   }

   close_lists();
   
   time( &stop);
   printf( "Start %s\n", asctime(localtime(&start)) );
   printf( "Stop  %s\n", asctime(localtime(&stop)) );
   printf( "Songs = %i  Albums = %i  Artists = %i  Podcasts = %i  Links = %i\n", song_cnt, album_cnt, artist_cnt, podcast_cnt, link_cnt );
   printf( "Total Lines = %i  Tracks = %i\n", lines, track_cnt );

   done();
   return(0);
}

int done() {

   printf( "Press any key to continue !\n" );
   getc( stdin );
   exit(0);
}

int close_lists() {
	int i;
	LIST * cur;



   write_list( "albums.txt",    albums );
   write_list( "artists.txt",   artists );
   write_list( "songs.txt",     songs );
   write_list( "playlists.txt", playlists );
   write_list( "podcasts.txt",  podcasts );
   write_link( "links.txt",     linksx );
   
   Close_Artwork();

}

int write_list( char * file, LIST * list ) {
int i;
LIST * cur;
FILE * out;

   if ( list ) {
      out = fopen( file, "w" );
      cur = list;
      while ( cur ) {
         for ( i=0; i<cur->trkcnt; i++ ) {
            if ( !i ) fprintf( out, "%s=%i", cur->name, cur->tracks[i] );
            else fprintf( out, "\t%i", cur->tracks[i] );
         }
         fprintf( out, "\n" );
         cur = cur->next;
      }
      fclose( out );
   }
   return(0);
}

int write_link( char * file, LIST * list ) {
int i=0;
char buf[500];
LIST * cur;
FILE * out, * lnk;

   if ( list ) {
      out = fopen( "bld_links", "w" );
      lnk = fopen( file, "w" );
      getcwd( buf, 500 );
      fprintf( out, "cd \"%s\"\n", buf );
      cur = list;
      while ( cur ) {
         strcpy( buf, cur->name );
         fprintf( lnk, "%s\n", buf );
         decode_lnk( buf );
         fprintf( out, "ln -s \"/%s\" link%d\n", buf, i++ );
         cur = cur->next;
      }
      fclose( out );
      fclose( lnk );
		chmod( "bld_links", 0777 );
   }
   return(0);
}

int insert_trk() {
int i, j, k, l, x=0, podcast=0, vid=0;
char *id=0, *artist=0, *album=0, *song=0, *loc=0, *lnk, buf[500];

   for ( i=0; i<keys; i++ ) {
      if ( !strcmp( key[i].name, "Name"   ))      song    = key[i].value;
      if ( !strcmp( key[i].name, "Album"  ))      album   = key[i].value;
      if ( !strcmp( key[i].name, "Artist" ))      artist  = key[i].value;
      if ( !strcmp( key[i].name, "Location" ))    loc     = key[i].value;
      if ( !strcmp( key[i].name, "Track ID" ))    id      = key[i].value;
      if ( !strcmp( key[i].name, "Playlist ID" )) id      = key[i].value;
      if ( !strcmp( key[i].name, "Podcast" ))     podcast = 1;
      if ( !strcmp( key[i].name, "Has Video" ) ||
           !strcmp( key[i].name, "Video" )     ||
           !strcmp( key[i].name, "Movie" ) )      vid     = 1;
   }

   if ( !album ) {
      i = keys++;
      album = "Unknown Album";
      key[i].name  = (char*)malloc( strlen("Album")+8 );
      key[i].value = (char*)malloc( strlen(album)+8 );
      strcpy( key[i].name,  "Album" );
      strcpy( key[i].value, album );
   }
   if ( !artist ) {
      i = keys++;
      artist = "Unknown Artist";
      key[i].name  = (char*)malloc( strlen("Artist")+8 );
      key[i].value = (char*)malloc( strlen(artist)+8 );
      strcpy( key[i].name,  "Artist" );
      strcpy( key[i].value, artist );
   }
   if ( !song ) {
      i = keys++;
      song = "Unknown Name";
      key[i].name  = (char*)malloc( strlen("Name")+8 );
      key[i].value = (char*)malloc( strlen(song)+8 );
      strcpy( key[i].name,  "Name" );
      strcpy( key[i].value, song );
   }

   x = atoi( id );
   if ( !x ) return(0);

   if ( podcast ) {
      if ( insert_list( (LIST*)&podcasts,  album,  x, 1 )) podcast_cnt++;
   } else {
      if ( !vid ) {
         if ( insert_list( (LIST*)&albums,  album,  x, 1 )) album_cnt++;
         if ( insert_list( (LIST*)&artists, artist, x, 1 )) artist_cnt++;
         if ( insert_list( (LIST*)&songs,   song,   x, 1 )) song_cnt++;
      }
   }

   if ( loc ) {
      if ( !strncmp( loc, "file://localhost/", 17 )) {
         lnk = (char*)malloc( strlen( loc )+8 );
         strcpy( lnk, loc );
         i = 17;
         l = 0;
         for ( j=i; j<strlen(lnk); j++ ) {
            if ( lnk[j] == '/' ) {
               l++;
               for ( k=j+1; k<strlen(lnk); k++ ) {
                  if ( lnk[k] == '/' ) break;
                  if ( lnk[k] == '%' && lnk[k+3] == '%' ) lnk[j+1] = 0;
               }
            }
         }
         k = 0;
         if ( l > 5 ) l = 3; else l = 1;
         for ( j=strlen(lnk); j>i; j-- ) {
            if ( lnk[j] == '/' ) {
               if ( ++k == l ) break;
            }
         }
         lnk[j+1] = 0;
         if ( insert_list( (LIST*)&linksx, &lnk[i], x, 0 )) {
            link_cnt++;
         } else {
            free( lnk );
         }
      }
      if ( !vid ) {
         Locate_Art( album, artist, song, loc );
      }
   }

   sprintf( buf, "Track-%s.txt", id );
   out = fopen( buf, "w" );
   for ( i=0; i<keys; i++ ) {
      fprintf( out, "%s=%s\t", key[i].name, key[i].value );
   }
   fclose( out );
   
   if ( (++track_cnt%100) == 0 ) printf( "Track count = %i\n", track_cnt );

	return(0);
}

int insert_play() {
int i, f, x=0;
char *name=0, *b, buf[200];

   for ( i=0; i<keys; i++ ) {
      if ( !strcmp( key[i].name, "Name"   )) name = key[i].value;
   }

   if ( !strcmp( name, "Library" ) || !strcmp( name, "Music" )) name = 0;
   if ( name ) playlist_cnt++;
   
   while ( 1 ) {
      b = fgets( buf, 200, xml );
      if ( !b ) break;
      lines++;

      f = getflds( b );
      if ( !name ) f = 3;
      switch ( f ) {
         case 0 :
                  if ( !strcmp( flds[0], "/array" )) return(0);
                  break;
         case 1 :
                  return(0);
         case 2 :
                  if ( strcmp( flds[1], "Track ID" )) return(0);
                  x = atoi( flds[2] );
                  if ( !x ) return(0);
                  insert_list( (LIST*)&playlists, name,  x, 1 );
         case 3 :
                  if ( !strcmp( flds[0], "/array" )) return(0);
                  break;
      }
   }

	return(0);	
}

int insert_list( LIST * list, char * name, int track, int flag ) {
int i, x;
LIST *prev=0, *cur, *item;

   if ( !list->next ) {
      list->next = acq_item( name, track );
      return(1);
   }
   cur = (LIST*)list->next;
   while ( cur ) {
      i = strcmp( cur->name, name );
      if ( !i ) {
         if ( flag ) {
            x = cur->trkcnt++;
            if ( x < 10 ) cur->tracks = (int*)realloc( cur->tracks, sizeof(int)*(x+8) );
            else cur->tracks = (int*)realloc( cur->tracks, sizeof(int)*(((x+8)+99)/100*100) );
            if ( !cur->tracks ) {
               printf( "* Out of memory !\n" );
               exit( 1 );
            }
            cur->tracks[x] = track;
         }
         return(0);
      }
      if ( i > 0 ) break;
      prev = cur;
      cur = (LIST*)cur->next;
   }

   item = acq_item( name, track );
   item->next = cur;
   if ( !prev ) list->next = item;
   else prev->next = item;
   return(1);
}

LIST * acq_item( char * name, int track ) {
LIST * item;

   item = (LIST*)calloc( sizeof(LIST), 1 );
   item->name = name;
   item->trkcnt = 1;
   item->tracks = (int*)malloc( sizeof(int) );
   item->tracks[0] = track;
   
   return( item );
}

int getflds( char * in ) {
int i, f=0, str[4], end[4];

   for ( i=0; i<4; i++ ) { str[i] = end[i] = 0; }
   i = 0;
   while ( 1 ) {
      while ( in[i] && in[i] != '<' ) i++;
      if ( !in[i] ) break;
      
      str[0] = ++i;                          // start of key
      while ( in[i] && in[i] != '>' ) i++;
      if ( !in[i] ) break;
      end[0] = i;
      in[i++] = 0;

      str[1] = i;                            // label
      while ( in[i] && in[i] != '<' ) i++;
      if ( !in[i] ) break;
      end[1] = i;
      in[i++] = 0;

      while ( in[i] && in[i] != '>' ) i++;   // skip over type
      if ( !in[i] ) break;
      i++;
      while ( in[i] && in[i] != '>' ) i++;
      if ( !in[i] ) break;

      str[2] = ++i;                          // value
      while ( in[i] && in[i] != '<' ) i++;
      if ( !in[i] ) break;
      end[2] = i;
      in[i] = 0;
   }

   flds[0] = flds[1] = flds[2] = flds[3] = dummy;

   if ( end[0] ) { flds[0] = &in[str[0]]; f = 0; }
   if ( end[1] ) { flds[1] = &in[str[1]]; f = 1; }
   if ( end[2] ) { flds[2] = &in[str[2]]; f = 2; }
   
   return(f);
    
}

int decode_txt( char * in ) {
int i;
char p[1000];

   utf8_decode( in, p );

   for ( i=0; i<strlen(p); i++ ) {
      if ( !p[i] ) break;
      if ( p[i] == '&' && p[i+1] == '#' && p[i+4] == ';' ) {
         p[i] = (p[i+2]-'0') * 10 + (p[i+3]-'0');
         strcpy( &p[i+1], &p[i+5] );
      }
   }

	return(0);
}

int decode_lnk( char * in ) {
int i, j;
char c, *p;

   p = in;
   for ( i=0; i<strlen(p); i++ ) {
      if ( !p[i] ) break;
      if ( p[i] == '%' ) {
         c = p[i+1];
         if ( c < 'A' ) j = c - '0'; else j = c - 'A' + 10;
         j = j * 16;
         c = p[i+2];
         if ( c < 'A' ) j += c - '0'; else j += c - 'A' + 10;
         p[i] = j;
         strcpy( &p[i+1], &p[i+3] );
      }
   }

	return(0);
}

char * decode_url( char * in ) {
int i, j, k, x, y;
char *p, *q;

   p = in;
   for ( i=0; i<strlen(p); i++ ) {
      if ( !p[i] ) break;
      if ( p[i] == '&' && p[i+1] == '#' && p[i+4] == ';' ) {
         j = (p[i+2]-'0') * 10 + (p[i+3]-'0');
         x = j / 16;
         y = j % 16;
         if ( x > 9 ) x += 'a'-10; else x += '0';
         if ( y > 9 ) y += 'a'-10; else y += '0';
         p[i++] = '%';
         p[i++] = x;
         p[i++] = y;
         strcpy( &p[i], &p[i+2] );
      }
   }
   p[i] = 0;

   for ( i=0; i<strlen(p); i++ ) { // Get rid of utf8 encode
      if ( !p[i] ) break;
      if ( p[i] == '%' && p[i+1] == 'C' && p[i+3] == '%' ) {
         j = (Hex2Bin(p[i+1]) << 4) | Hex2Bin(p[i+2]);
         k = (Hex2Bin(p[i+4]) << 4) | Hex2Bin(p[i+5]);
         j = (( j & 3 ) << 6 ) | ( k & 077 );
         x = j / 16;
         y = j % 16;
         if ( x > 9 ) x += 'a'-10; else x += '0';
         if ( y > 9 ) y += 'a'-10; else y += '0';
         p[i++] = '%';
         p[i++] = x;
         p[i] = y;
         strcpy( &p[i+1], &p[i+4] );
      }
   }
   p[i] = 0;

   q = (char*)malloc( strlen(p) * 2 );
   j = 0;
   for ( i=0; i<strlen(p); i++ ) {
      if ( p[i] == 39 ) {
         strncpy( &q[j], "%27", 3 );
         j += 3;
      } else q[j++] = p[i];
   }
   q[j] = 0;

   p = realloc( p, strlen(q) + 20 );
   strcpy( p, q );
   free( q );

   return(p);
}

int Locate_Art( char * album, char * artist, char * song, char * place ) {
int i, j, x;
DIR  * dir;
struct stat sta;
struct dirent * ent;
char Album[200], Artist[200], Song[200];
char * path, * name, loc[1000], buf[1000];



   Decode( place, loc );
   Decode( album, Album );
   Decode( artist, Artist );
   Decode( song, Song );
   
   if ( !strncasecmp( loc, "file://localhost/", 17 )) path = &loc[17]; else path = loc;
   i = strlen( loc );
   j = 0;
   while ( --i ) {
      if ( !j && loc[i] == '.' ) { loc[i] = 0; j++; continue; }
      if ( loc[i] == '/' ) { loc[i] = 0; name = &loc[i+1]; break; }
   }

   dir = opendir( path );
   if ( dir ) {
      while ( (ent=readdir(dir) ) ) {
         i = strlen( ent->d_name );
         x = 0;
         if ( i > 4 && !strcasecmp( &ent->d_name[i-4], ".jpg" )) {
            sprintf( buf, "%s/%s", path, ent->d_name );
            if ( !strncasecmp( ent->d_name, "folder", 6 ) ||
                 !strncasecmp( ent->d_name, "cover", 5 ) ) {
               Ins_Artwork( 0, Album, buf );
               Ins_Artwork( 1, Artist, buf );
               Ins_Artwork( 2, Song, buf );
               x++;

            } else {
               if ( !strncasecmp( ent->d_name, "albumart", 8 ) ) {
                  if ( !Scan( ent->d_name, "large" ) ) {
                     Ins_Artwork( 0, Album, buf );
                     Ins_Artwork( 1, Artist, buf );
                     Ins_Artwork( 2, Song, buf );
                  x++;

                  }
               }
            }
            if ( !x ) {
               if ( !Scan( ent->d_name, Song )) {
                  Ins_Artwork( 2, Song, buf );

               }
               if ( !Scan( ent->d_name, Artist )) {
                  Ins_Artwork( 1, Artist, buf );

               }
               if ( !Scan( ent->d_name, Album )) {
                  Ins_Artwork( 0, Album, buf );

               }
            }
         }
      }
      closedir( dir );
   }

   return(0);
}

int Ins_Artwork( int where, char * who,  char * path ) {
int i;
ART *prv=0, *cur, *nxt=0, *beg, *item;
char whom[1000];


   for ( i=0; i<strlen(who); i++ ) whom[i] = tolower( who[i] );
   whom[i] = 0;
   i = strlen( whom );
   while ( i && whom[i-1] == ' ' ) whom[i--] = 0;
   
   switch ( where ) {
      case 0 : beg = (ART*)&album_art;  break;
      case 1 : beg = (ART*)&artist_art; break;
      case 2 : beg = (ART*)&music_art;  break;
   }
   
   cur = beg->next;
   while ( cur ) {
      i = strcasecmp( cur->name, whom );

      if ( !i ) return(0);
      if ( i < 1 ) {
         prv = cur;
         cur = cur->next;
         continue;
      }
      break;
   }

   item = (ART*)calloc( sizeof(ART), 1 );
   item->name = malloc( strlen(whom)+1 );
   strcpy( item->name, whom );
   item->path = malloc( strlen(path)+1 );
   strcpy( item->path, path );

   if ( !prv ) beg->next = item;
   else prv->next = item;
   item->next = cur;

   art_cnt++;

   return(0);
}

int Close_Artwork() {
int i;
char buf[1000];
FILE * out;
ART *cur, *nxt=0, *beg;

   for ( i=0; i<3; i++ ) {
      switch ( i ) {
        case 0 :  beg = (ART*)&album_art;
                  if ( beg->next ) {

                     out = fopen( "album_art.txt", "w" );
                     break;
                  }
                  continue;
        case 1 :  beg = (ART*)&artist_art;
                  if ( beg->next ) {

                     out = fopen( "artist_art.txt", "w" );
                     break;
                  }
                  continue;
        case 2 :  beg = (ART*)&music_art;
                  if ( beg->next ) {

                     out = fopen( "music_art.txt", "w" );
                     break;
                  }
                  continue;
      }

      cur = beg->next;
      beg = 0;
      while ( cur ) {

         fprintf( out, "%s\t%s\n", cur->name, cur->path );
         
         nxt = cur->next;
         free( cur->name );
         free( cur->path );
         free( cur );
         
         cur = nxt;
      }
      fclose( out );

   }
   
   return(0);
   
}

int Decode( char * in, char * out ) {
int i, j=0, k, x, y;
char c, val[] = "0123456789ABCDEF ";
char * fld[100], ina[1000], buf[1000];
	
   strcpy( ina, in );
   
   for ( i=0; i<strlen(ina); i++ ) {
      if ( ina[i] == '%' ) {
         c = ina[++i];
         for ( k=0; k<17; k++ ) if ( c == val[k] || toupper(c) == val[k] ) break;
         if ( k > 15 ) {
            k = 0;
         }
         x = k << 4;
         c = ina[++i];
         for ( k=0; k<17; k++ ) if ( c == val[k] || toupper(c) == val[k] ) break;
         if ( k > 15 ) {
            k = 0;
         }
         buf[j++] = x + k;
      } else {
         buf[j++] = ina[i];
      }
   }
   buf[j] = 0;
   
   utf8_decode( buf, buf );
	
   x = 0;
   fld[x] = buf;
   j = strlen( buf );
   for ( i=0; i<j; i++ ) {
      if ( buf[i] == '/' ) {
         fld[++x] = &buf[i+1];
         buf[i] = 0;
      }
   }
   fld[++x] = NULL;
   i = x = 0;
   while ( fld[x] ) {
      k = 0;
      for ( j=0; j<strlen(fld[x]); j++ ) if ( fld[x][j] == ' ' ) { k = 0; break; }
      if ( k ) i += sprintf( &out[i], "\"%s\"/", fld[x++] );
      else i += sprintf( &out[i], "%s/", fld[x++] );
   }
   out[strlen(out)-1] = 0;
	
   return(0);
}

int utf8_decode( char * in, char * out ) {
int i=0, j, x=0;
unsigned char c1, c2, c3, data[2000];
	
	strcpy( data, in );
	j = strlen( in );
	
	while ( i < j ) {
		c1 = data[i];
		if ( c1 < 0x80 ) {
			out[x++] = c1;
			i++;
		} else if ( (c1 >= 0xC0) && (c1 <= 0xC3) ) {
			c2 = data[i+1];
			out[x++] = ((c1 & 037) << 6) | (c2 & 077);
			i += 2;
		} else {
			out[x++] = c1;
			i++;
		}
	}
	out[x] = 0;
	
   return(0);
}

int Scan( char * fld, char * val ) {
	int i, j, l;
	
   l = strlen( val );
   j = strlen( fld ) - l + 1;
   for ( i=0; i<j; i++ ) {
      if ( !strncasecmp( &fld[i], val, l )) return(0);
   }
	
   return(1);
	
}

int Hex2Bin( char in ) {
	char c;
	
   c = tolower(in);
   if ( c >= 'a' ) return( c - 'a' + 10 ); else return( c - '0' );
}









